# -*- coding: utf-8 -*-

"""Bio2BEL is a project aimed at bringing biological databases and other structured knowledge sources in a wide variety
of schemata into the unified framework of `Biological Expression Language <http://http://openbel.org/>`_.

This package provides guidelines, tutorials, and tools for making standardized ``bio2bel`` packages as well as a
unifying framework for integrating them.

Installation
------------
Easiest
~~~~~~~
Download the latest stable code from `PyPI <https://pypi.org/bio2bel>`_ with:

.. code-block:: sh

   $ python3 -m pip install bio2bel

Get the Latest
~~~~~~~~~~~~~~~
Download the most recent code from `GitHub <https://github.com/bio2bel/bio2bel>`_ with:

.. code-block:: sh

   $ python3 -m pip install git+https://github.com/bio2bel/bio2bel.git

For Developers
~~~~~~~~~~~~~~
Clone the repository from `GitHub <https://github.com/bio2bel/bio2bel>`_ and install in editable mode with:

.. code-block:: sh

   $ git clone https://github.com/bio2bel/bio2bel.git
   $ cd bio2bel
   $ python3 -m pip install -e .


Testing
-------
Bio2BEL is tested with Python3 on Linux using `Travis CI <https://travis-ci.org/bio2bel/bio2bel>`_.
"""

from pkg_resources import iter_entry_points

from . import abstractmanager, cli_utils, utils
from .abstractmanager import *
from .cli_utils import *
from .downloading import *
from .utils import *

GROUP_NAME = 'bio2bel'

# Allow `from bio2bel.ext import *`
entries = [
    entry_point.name
    for entry_point in iter_entry_points(group=GROUP_NAME, name=None)
]


def setup():
    """Add the :code:`bio2bel.ext` importer/loader to the meta_path. Should probably only be called once."""
    from .exthook import ExtensionImporter
    importer = ExtensionImporter(GROUP_NAME)
    importer.install()


setup()
del setup  # Do a bit of cleanup. Not sure if it's necessary, but Flask did...

__version__ = '0.0.10'

__title__ = 'bio2bel'
__description__ = "A package for converting biological stuff to BEL"
__url__ = 'https://github.com/bio2bel/bio2bel'

__author__ = 'Charles Tapley Hoyt'
__email__ = 'charles.hoyt@scai.fraunhofer.de'

__license__ = 'Apache 2.0 License'
__copyright__ = 'Copyright (c) 2017-2018 Charles Tapley Hoyt'

__all__ = entries + abstractmanager.__all__ + utils.__all__ + cli_utils.__all__



# -*- coding: utf-8 -*-

import unittest

from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base

import tests.constants
from bio2bel import AbstractManager
from bio2bel.exc import Bio2BELMissingNameError, Bio2BELModuleCaseError
from bio2bel.models import Action
from bio2bel.testing import AbstractTemporaryCacheClassMixin, MockConnectionMixin, TemporaryConnectionMixin
from tests.constants import NUMBER_TEST_MODELS

class TestManagerFailures(unittest.TestCase):
    """Test improperly implement AbstractManager."""

    def test_missing_all_abstract(self):
        """Test that the abstract class can't be instantiated."""

        class Manager(AbstractManager):
            """An incompletely implement AbstractManager"""

        with self.assertRaises(TypeError):  # cant's instantiate abstract class
            Manager()

    def test_fail_instantiation_2(self):
        """Test that the abstract class can't be instantiated."""
        base = declarative_base()

        class Manager(AbstractManager):
            """An incompletely implement AbstractManager"""

            @property
            def _base(self):
                return base

        with self.assertRaises(TypeError):
            Manager()

    def test_fail_instantiation_3(self):
        """Test that the abstract class can't be instantiated."""

        class Manager(AbstractManager):
            """An incompletely implement AbstractManager."""

            def populate(self):
                """Populate the database."""

        with self.assertRaises(TypeError):
            Manager()

    def test_undefined_module_name(self):
        """Test error thrown if module name isn't set."""
        base = declarative_base()

        class Manager(AbstractManager):
            """An improperly implemented AbstractManager that is missing the module_name class variable."""

            @property
            def _base(self):
                return base

            def is_populated(self):
                """Check if the database is already populated."""

            def populate(self):
                """Populate the database."""

        with self.assertRaises(Bio2BELMissingNameError):
            Manager()

    def test_module_name_case(self):
        """Test error thrown if module name is weird case."""
        base = declarative_base()

        class Manager(AbstractManager):
            module_name = 'TESTOMG'

            @property
            def _base(self):
                return base

            def is_populated(self):
                """Check if the database is already populated."""

            def populate(self):
                """Populate the database."""

        with self.assertRaises(Bio2BELModuleCaseError):
            Manager()


class TestManagerEnsure(TemporaryConnectionMixin):
    """Tests the ensure function of an AbstractManager works properly."""

    def test_type_error(self):
        """Test that the connection's type is checked properly."""
        with self.assertRaises(TypeError):
            tests.constants.Manager.ensure(connection=5)

    def test_build(self):
        """Test that a string can be passed."""
        m = tests.constants.Manager.ensure(connection=self.connection)
        self.assertIsInstance(m, tests.constants.Manager)

    def test_pass_through(self):
        """Test that a manager can be passed."""
        m = tests.constants.Manager(connection=self.connection)
        self.assertIsInstance(m, tests.constants.Manager)

        m2 = tests.constants.Manager.ensure(connection=m)
        self.assertIsInstance(m, tests.constants.Manager)
        self.assertEqual(m, m2)


class TestConnectionDropping(MockConnectionMixin, AbstractTemporaryCacheClassMixin):
    """Tests dropping the database."""
    Manager = tests.constants.Manager

    def test_no_exist(self):
        """Check if the database gets dropped that stuff breaks."""

        with self.mock_global_connection:  # don't want to worry about that drop_app hook
            self.assertEqual(0, Action.count())
            self.manager.drop_all()
            self.assertEqual(1, Action.count())

        with self.assertRaises(OperationalError):
            self.manager.get_model_by_model_id(5)


class TestConnectionLoading(AbstractTemporaryCacheClassMixin):
    """Tests the connection is loaded properly."""

    Manager = tests.constants.Manager

    def test_connection(self):
        """Tests the type of the connection"""
        self.assertIsNotNone(self.connection)
        self.assertIsInstance(self.connection, str)

    def test_manager_passes(self):
        """Test the connection is inside the manager properly"""
        self.assertEqual(self.connection, self.manager.connection)

    def test_repr(self):
        """Test the repr function of the AbstractManager."""
        self.assertEqual('<TestManager url={}>'.format(self.connection), repr(self.manager))

    def test_get_missing_model(self):
        """Test population."""
        self.assertFalse(self.manager.is_populated())
        self.manager.populate()
        self.assertTrue(self.manager.is_populated())

        self.assertEqual(NUMBER_TEST_MODELS, self.manager.count_model())

        self.assertIsNone(self.manager.get_model_by_model_id(0))
        self.assertIsNone(self.manager.get_model_by_model_id(1))
        self.assertIsNone(self.manager.get_model_by_model_id(2))
        self.assertIsNone(self.manager.get_model_by_model_id(3))
        self.assertIsNone(self.manager.get_model_by_model_id(4))
        self.assertIsNone(self.manager.get_model_by_model_id(150))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import logging
import os
import sys

import click

from .utils import get_data_dir

__all__ = [
    'add_cli_populate',
    'add_cli_drop',
    'add_cli_flask',
    'add_cli_summarize',
    'add_cli_to_bel',
    'add_cli_to_bel_namespace',
    'add_cli_cache',
]
log = logging.getLogger(__name__)


def add_cli_populate(main):
    """Add a ``populate`` command to main :mod:`click` function.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.option('--reset', is_flag=True)
    @click.pass_obj
    def populate(manager, reset):
        """Populates the database"""

        if reset:
            log.info('Deleting the previous instance of the database')
            manager.drop_all()
            log.info('Creating new models')
            manager.create_all()

        manager.populate()

    return main


def add_cli_drop(main):
    """Add a ``drop`` command to main :mod:`click` function.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.option('-y', '--yes', is_flag=True)
    @click.pass_obj
    def drop(manager, yes):
        """Drops database"""
        if yes or click.confirm('Drop everything?'):
            manager.drop_all()

    return main


def add_cli_flask(main):
    """Add a ``web`` comand main :mod:`click` function.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.option('-v', '--debug', is_flag=True)
    @click.option('-p', '--port')
    @click.option('-h', '--host')
    @click.pass_obj
    def web(manager, debug, port, host):
        """Run the web app."""
        app = manager.get_flask_admin_app(url='/')
        app.run(debug=debug, host=host, port=port)

    return main


def add_cli_to_bel(main):
    """Add several command to main :mod:`click` function related to export to BEL.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.option('-o', '--output', type=click.File('w'), default=sys.stdout)
    @click.pass_obj
    def to_bel(manager, output):
        """Write as BEL Script."""
        from pybel import to_bel
        graph = manager.to_bel()
        to_bel(graph, output)

    @main.command()
    @click.option('-c', '--connection')
    @click.pass_obj
    def upload_bel(manager, connection):
        """Upload BEL to network store."""
        from pybel import to_database
        graph = manager.to_bel()
        to_database(graph, connection=connection)

    return main


def add_cli_to_bel_namespace(main):
    """Add a ``upload_bel_namespace`` command to main :mod:`click` function.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.pass_obj
    def upload_bel_namespace(manager):
        """Upload names/identifiers to terminology store."""
        namespace = manager.upload_bel_namespace()
        click.echo('uploaded [{}] {}'.format(namespace.id, namespace))

    return main


def add_cli_summarize(main):
    """Add a ``summarize`` command to main :mod:`click` function.

    :param main: A click-decorated main function
    """

    @main.command()
    @click.pass_obj
    def summarize(manager):
        """Summarize the contents of the database."""
        for name, count in sorted(manager.summarize().items()):
            click.echo('{}: {}'.format(name.capitalize(), count))

    return main


def add_cli_cache(main):
    """Add several commands to main :mod:`click` function for handling the cache.

    :param main: A click-decorated main function
    """

    @main.group()
    def cache():
        pass

    @cache.command()
    @click.pass_obj
    def ls(manager):
        """Lists files in the cache."""
        data_dir = get_data_dir(manager.module_name)

        for path in os.listdir(data_dir):
            click.echo(path)

    @cache.command()
    @click.pass_obj
    def clear(manager):
        """Clears all files from the cache."""
        data_dir = get_data_dir(manager.module_name)

        for path in os.listdir(data_dir):
            if path in {'config.ini', 'cache.db'}:
                continue
            os.remove(os.path.join(data_dir, path))

    return main



# -*- coding: utf-8 -*-

"""Testing constants and utilities for Bio2BEL."""

import logging

from click.testing import CliRunner

import pybel
from bio2bel.namespacemanagermixin import Bio2BELMissingNamespaceModelError, NamespaceManagerMixin
from bio2bel.testing import AbstractTemporaryCacheMethodMixin, MockConnectionMixin, TemporaryConnectionMethodMixin
from pybel.manager.models import Base as PyBELBase, Namespace, NamespaceEntry
from tests.constants import Manager, Model, NUMBER_TEST_MODELS, TEST_MODEL_ID_FORMAT, TEST_MODEL_NAME_FORMAT

log = logging.getLogger(__name__)


class NamespaceManager(Manager, NamespaceManagerMixin):
    """Use parts of the test manager and finish the abstract namespace manager"""

    namespace_model = Model

    # automate by defining identifier column?

    def _create_namespace_entry_from_model(self, model, namespace=None):
        return NamespaceEntry(name=model.name, identifier=model.model_id, namespace=namespace)

    def _get_identifier(self, model):
        return model.model_id


class TestFailure(TemporaryConnectionMethodMixin):

    def test_type_failure(self):
        class _TestManager(Manager, NamespaceManagerMixin):
            def _create_namespace_entry_from_model(self, model, namespace=None):
                return NamespaceEntry(name=model.name, identifier=model.model_id, namespace=namespace)

            def _get_identifier(self, model):
                return model.model_id

        with self.assertRaises(Bio2BELMissingNamespaceModelError):
            _TestManager(connection=self.connection)

    def test_instantiation_failure(self):
        class _TestManager(Manager, NamespaceManagerMixin):
            """Use parts of the test manager and finish the abstract namespace manager"""

            namespace_model = Model

        with self.assertRaises(TypeError):
            _TestManager(connection=self.connection)


class TestNamespaceManagerMixin(TemporaryConnectionMethodMixin):
    def test_instantiation_success(self):
        """Test instantiation is possible."""
        NamespaceManager(connection=self.connection)


class TestAwesome(AbstractTemporaryCacheMethodMixin):
    Manager = NamespaceManager

    def setUp(self):
        super().setUp()
        PyBELBase.metadata.create_all(self.manager.engine, checkfirst=True)

    def populate(self):
        self.manager.populate()

    def test_namespace_name(self):
        self.assertEqual('test', self.manager.module_name)  # this is defined in the tests
        self.assertEqual('_TEST', self.manager._get_namespace_keyword())

    def test_make_namespace(self):
        namespace = self.manager._make_namespace()
        self.assertIsNotNone(namespace)
        self.assertIsInstance(namespace, Namespace)
        self.assertEqual('_TEST', namespace.keyword)
        self.assertEqual('_TEST', namespace.url)

        self.assertEqual(NUMBER_TEST_MODELS, namespace.entries.count())

        em = {
            entry.identifier: entry
            for entry in namespace.entries
        }

        for i in range(NUMBER_TEST_MODELS):
            model_id = TEST_MODEL_ID_FORMAT.format(i)

            self.assertIn(model_id, em)
            model = em[model_id]

            self.assertEqual(model_id, model.identifier)
            self.assertEqual(TEST_MODEL_NAME_FORMAT.format(i), model.name)
            self.assertEqual(namespace, model.namespace)

    def test_update_namespace(self):
        self.manager._make_namespace()

        # mock some sort of changes to the database

        _number_to_add = 4

        models = [
            Model.from_id(model_id)
            for model_id in range(NUMBER_TEST_MODELS + 1, NUMBER_TEST_MODELS + 1 + _number_to_add)
        ]
        self.manager.session.add_all(models)
        self.manager.session.commit()

        namespace = self.manager._update_namespace()

        self.assertIsNotNone(namespace)
        self.assertIsInstance(namespace, Namespace)
        self.assertEqual(NUMBER_TEST_MODELS + _number_to_add, namespace.entries.count())


class TestCli(MockConnectionMixin):
    """Tests the CLI for uploading a BEL namespace."""

    def setUp(self):
        """Set up a CliRunner and an accompanying CLI for each test."""
        self.runner = CliRunner()
        self.main = NamespaceManager.get_cli()
        self.manager = Manager(connection=self.connection)
        self.manager.populate()

    def test_to_bel_namespace(self):
        """Test the population function can be run."""
        self.assertEqual(5, self.manager.count_model(), msg='manager should be populated')

        pybel_manager = pybel.Manager(connection=self.connection)
        self.assertEqual(0, pybel_manager.count_namespaces())

        args = [
            '--connection',
            self.connection,
            'upload_bel_namespace'
        ]
        self.runner.invoke(self.main, args)

        self.assertEqual(1, pybel_manager.count_namespaces())



# -*- coding: utf-8 -*-

"""Provide abstractions over BEL namespace generation procedures"""

import logging
import time
from abc import abstractmethod

from sqlalchemy import and_
from tqdm import tqdm

from pybel.manager.models import Namespace
from .abstractmanager import AbstractManager
from .cli_utils import add_cli_to_bel_namespace

log = logging.getLogger(__name__)

__all__ = [
    'Bio2BELMissingNamespaceModelError',
    'NamespaceManagerMixin',
]


class Bio2BELMissingNamespaceModelError(TypeError):
    """Raised when the namespace_model class variable is not defined"""


class NamespaceManagerMixin(AbstractManager):
    """This mixin adds functions for making a BEL namespace to a repository

    *How to Use This Mixin*

    1. Either include it as a second inheriting class after :class:`AbstractManager` (this is how mixins are usually
    used):

    ..code-block:: python


        from bio2bel import AbstractManager
        from bio2bel.namespacemanagermixin import NamespaceManagerMixin

        class MyManager(AbstractManager, NamespaceManagerMixin):
            ...


    1. Or subclass it directly, since it also inherits from :class:`AbstractManager`, like:

    ..code-block:: python

        from bio2bel.namespacemanagermixin import NamespaceManagerMixin

        class MyManager(NamespaceManagerMixin):
            ...

    """
    namespace_model = ...

    def __init__(self, *args, **kwargs):
        """
        :param Optional[str] connection: SQLAlchemy connection string
        """
        if self.namespace_model is ...:
            raise Bio2BELMissingNamespaceModelError

        super().__init__(*args, **kwargs)

    @abstractmethod
    def _create_namespace_entry_from_model(self, model, namespace=None):
        """"""

    @abstractmethod
    def _get_identifier(self, model):
        """Given an instance of namespace_model, extract its identifier"""

    def _get_namespace_entries(self):
        return [
            self._create_namespace_entry_from_model(model)
            for model in self._iterate_namespace_models()
        ]

    def _iterate_namespace_models(self):
        """Return an iterator over the models to be converted to the namespace"""
        return tqdm(self._get_query(self.namespace_model), total=self._count_model(self.namespace_model))

    @classmethod
    def _get_namespace_keyword(cls):
        """Gets the keyword to use as the reference BEL namespace.

        :rtype: str
        """
        return '_{}'.format(cls.module_name.upper())

    @classmethod
    def _get_namespace_filter(cls):
        """Get an SQLAlchemy filter for getting the reference BEL namespace.

        :return: A SQLAlchemy namespace filter
        """
        _namespace_keyword = cls._get_namespace_keyword()

        return and_(
            Namespace.keyword == _namespace_keyword,
            Namespace.url == _namespace_keyword
        )

    def _get_default_namespace(self):
        """Get the reference BEL namespace if it exists.

        :rtype: Optional[pybel.manager.models.Namespace
        """

        namespace_filter = self._get_namespace_filter()
        return self._get_query(Namespace).filter(namespace_filter).one_or_none()

    def _make_namespace(self):
        """
        :rtype: pybel.manager.models.Namespace
        """
        from pybel.manager.models import Namespace

        entries = self._get_namespace_entries()
        _namespace_keyword = self._get_namespace_keyword()
        ns = Namespace(
            name=_namespace_keyword,
            keyword=_namespace_keyword,
            url=_namespace_keyword,
            version=str(time.asctime()),
            entries=entries,
        )
        self.session.add(ns)

        t = time.time()
        log.info('committing models')
        self.session.commit()
        log.info('committed models in %.2f seconds', time.time() - t)

        return ns

    @staticmethod
    def _build_old_entry_set(namespace):
        """Converts PyBEL generalized namespace entries to a set.

        Default to using the identifier, but can be overridden to use the name instead.

        :rtype: set[pybel.manager.model.NamespaceEntry]

        >>> {term.identifier for term in namespace.entries}
        """
        return {term.identifier for term in namespace.entries}

    def _update_namespace(self):
        """Only call this if namespace won't be none!

        :rtype: pybel.manager.models.Namespace
        """
        namespace = self._get_default_namespace()

        old_entry_set = self._build_old_entry_set(namespace)
        new_count = 0

        for model in self._iterate_namespace_models():
            if self._get_identifier(model) in old_entry_set:
                continue

            new_count += 1
            entry = self._create_namespace_entry_from_model(model, namespace=namespace)
            self.session.add(entry)

        t = time.time()
        log.info('got %d new entries. committing models', new_count)
        self.session.commit()
        log.info('committed models in %.2f seconds', time.time() - t)

        return namespace

    def upload_bel_namespace(self):
        """
        :rtype: pybel.manager.models.Namespace
        """
        if not self.is_populated():
            self.populate()

        ns = self._get_default_namespace()

        if ns is None:
            return self._make_namespace()

        return self._update_namespace()

    @staticmethod
    def _cli_add_to_bel_namespace(main):
        return add_cli_to_bel_namespace(main)

    @classmethod
    def get_cli(cls):
        """Gets a :mod:`click` main function to use as a command line interface."""
        main = super().get_cli()
        cls._cli_add_to_bel_namespace(main)
        return main



# -*- coding: utf-8 -*-

"""Provides abstractions over the management of SQLAlchemy connections and sessions."""

import logging
from abc import ABCMeta, abstractmethod
from functools import wraps

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from .cli_utils import add_cli_cache, add_cli_drop, add_cli_flask, add_cli_populate, add_cli_summarize, add_cli_to_bel
from .exc import Bio2BELMissingModelsError, Bio2BELMissingNameError, Bio2BELModuleCaseError
from .models import Action, create_all
from .utils import get_connection

__all__ = [
    'AbstractManagerConnectionMixin',
    'AbstractManager',
]

log = logging.getLogger(__name__)


class AbstractManagerMeta(ABCMeta):
    """Crazy metaclass to hack in a decorator to the populate function."""

    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        cls._populate_original = cls.populate

        @wraps(cls._populate_original)
        def populate_wrapped(self, *populate_args, **populate_kwargs):
            """Populates the database."""
            cls._populate_original(self, *populate_args, **populate_kwargs)

            # Hack in the action storage
            self._store_populate()

        cls.populate = populate_wrapped

        return cls


class AbstractManagerConnectionMixin(object):
    """Represents the connection-building aspect of the abstract manager.

    Minimally requires the definition of the class-level variable, ``module_name``.

    Example for InterPro:

    >>> from bio2bel.abstractmanager import AbstractManagerConnectionMixin
    >>> class Manager(AbstractManagerConnectionMixin):
    >>>     module_name = 'interpro'


    In general, this class won't be used directly except in the situation where the connection should be loaded
    in a different way and it can be used as a mixin.
    """

    #: This represents the module name. Needs to be lower case
    module_name = ...

    def __init__(self, connection=None):
        """
        :param Optional[str] connection: SQLAlchemy connection string
        """
        if not self.module_name or not isinstance(self.module_name, str):
            raise Bio2BELMissingNameError('module_name class variable not set on {}'.format(self.__class__.__name__))

        if self.module_name != self.module_name.lower():
            raise Bio2BELModuleCaseError('module_name class variable should be lowercase')

        self.connection = self.get_connection(connection=connection)
        self.engine = create_engine(self.connection)
        self.session_maker = sessionmaker(bind=self.engine, autoflush=False, expire_on_commit=False)
        self.session = scoped_session(self.session_maker)

    @classmethod
    def get_connection(cls, connection=None):
        """Gets the default connection string by wrapping :func:`bio2bel.utils.get_connection` and passing
        this class's :data:`module_name` to it.

        :param Optional[str] connection: A custom connection to pass through
        :rtype: str
        """
        return get_connection(cls.module_name, connection=connection)

    def _store_populate(self):
        create_all(self.engine)
        Action.store_populate(self.module_name, session=self.session)

    def _store_drop(self):
        create_all(self.engine)
        Action.store_drop(self.module_name, session=self.session)

    def __repr__(self):
        return '<{module_name}Manager url={url}>'.format(
            module_name=self.module_name.capitalize(),
            url=self.engine.url
        )


class _FlaskMixin(AbstractManagerConnectionMixin):
    """Mixin for making the AbstractManager build a Flask application."""

    #: Represents a list of SQLAlchemy classes to make a Flask-Admin interface.
    flask_admin_models = ...

    def _add_admin(self, app, **kwargs):
        """Add a Flask Admin interface to an application.

        :param flask.Flask app: A Flask application
        :param kwargs: Keyword arguments are passed through to :class:`flask_admin.Admin`
        :rtype: flask_admin.Admin
        """
        if self.flask_admin_models is ...:
            raise Bio2BELMissingModelsError

        from flask_admin import Admin
        from flask_admin.contrib.sqla import ModelView

        admin = Admin(app, **kwargs)

        for flask_admin_model in self.flask_admin_models:
            if isinstance(flask_admin_model, tuple):  # assume its a 2 tuple
                if len(flask_admin_model) != 2:
                    raise TypeError

                model, view = flask_admin_model
                admin.add_view(view(model, self.session))

            else:
                admin.add_view(ModelView(flask_admin_model, self.session))

        return admin

    def get_flask_admin_app(self, url=None):
        """Create a Flask application if this class has defined the :data:`flask_admin_models` variable a list of
        model classes.

        :param Optional[str] url: Optional mount point of the admin application. Defaults to ``'/'``.
        :rtype: flask.Flask
        """
        from flask import Flask

        app = Flask(__name__)
        self._add_admin(app, url=(url or '/'))
        return app


class _QueryMixin(AbstractManagerConnectionMixin):
    def _get_query(self, model):
        """Gets a query for the given model using this manager's session.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :return: a SQLAlchemy query
        """
        return self.session.query(model)

    def _count_model(self, model):
        """Helps count the number of a given model in the database.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :rtype: int
        """
        return self._get_query(model).count()

    def _list_model(self, model):
        """Helps get all instances of the model in the database.

        :param sqlalchemy.ext.declarative.api.DeclarativeMeta model: A SQLAlchemy model class
        :rtype: list
        """
        return self._get_query(model).all()


class _CliMixin(AbstractManagerConnectionMixin):
    """Functions for building a CLI"""

    @classmethod
    def _get_cli_main(cls):
        """Build a :mod:`click` CLI main function.

        :param Type[AbstractManager] cls: A Manager class
        :return: The main function for click
        """
        import click

        @click.group(help='Default connection at {}'.format(cls.get_connection()))
        @click.option('-c', '--connection', help='Defaults to {}'.format(cls.get_connection()))
        @click.pass_context
        def main(ctx, connection):
            """Bio2BEL CLI."""
            logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            logging.getLogger('bio2bel.utils').setLevel(logging.WARNING)
            ctx.obj = cls(connection=connection)

        return main

    @staticmethod
    def _cli_add_populate(main):
        return add_cli_populate(main)

    @staticmethod
    def _cli_add_drop(main):
        return add_cli_drop(main)

    @staticmethod
    def _cli_add_cache(main):
        return add_cli_cache(main)

    @staticmethod
    def _cli_add_flask(main):
        return add_cli_flask(main)

    @staticmethod
    def _cli_add_to_bel(main):
        return add_cli_to_bel(main)

    @staticmethod
    def _cli_add_summarize(main):
        return add_cli_summarize(main)

    @classmethod
    def get_cli(cls):
        """Gets a :mod:`click` main function to use as a command line interface."""
        main = cls._get_cli_main()

        cls._cli_add_populate(main)
        cls._cli_add_drop(main)
        cls._cli_add_cache(main)

        if hasattr(cls, 'flask_admin_models') and cls.flask_admin_models:
            cls._cli_add_flask(main)

        if hasattr(cls, 'to_bel'):
            cls._cli_add_to_bel(main)

        if hasattr(cls, 'summarize'):
            cls._cli_add_summarize(main)

        return main


class AbstractManager(_FlaskMixin, _QueryMixin, _CliMixin, metaclass=AbstractManagerMeta):
    """This is a base class for implementing your own Bio2BEL manager.

    It already includes functions to handle configuration, construction of a connection to a database using SQLAlchemy,
    creation of the tables defined by your own :func:`sqlalchemy.ext.declarative.declarative_base`, and has hooks to
    override that populate and make simple queries to the database. Since :class:`AbstractManager` inherits from
    :class:`abc.ABC` and is therefore an abstract class, there are a few class variables, functions, and properties
    that need to be overridden.

    **Overriding the Module Name**

    First, the class-level variable ``module_name`` must be set to a string corresponding to the name of the data
    source.

    .. code-block:: python

        from bio2bel import AbstractManager

        class Manager(AbstractManager):
            module_name = 'mirtarbase'  # note: use lower case module names

    In general, this should also correspond to the same value as ``MODULE_NAME`` set in ``constants.py`` and can also
    be set with an assignment to this value

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME

        class Manager(AbstractManager):
            module_name = MODULE_NAME

    **Setting the Declarative Base**

    Building on the previous example, the (private) abstract property :data:`bio2bel.AbstractManager._base` must be
    overridden to return the value from your :func:`sqlalchemy.ext.declarative.declarative_base`. We chose to make this
    an instance-level property instead of a class-level variable so each manager could have its own information about
    connections to the database.

    As a minimal example:

    .. code-block:: python

        from bio2bel import AbstractManager
        from sqlalchemy.ext.declarative import declarative_base

        Base = declarative_base()

        class Manager(AbstractManager):
            module_name = 'mirtarbase'  # note: use lower case module names

            @property
            def _base(self):
                return Base


    In general, the models should be defined in a module called ``models.py`` so the ``Base`` can also be imported.

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self):
                return Base

    **Populating the Database**

    Deciding how to populate the database using your SQLAlchemy models is incredibly creative and can't be given a good
    example without checking real code. See the previously mentioned `implementation of a Manager
    <https://github.com/bio2bel/mirtarbase/blob/master/src/bio2bel_mirtarbase/manager.py>`_.

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self):
                return Base

            def populate(self):
                ...

    **Checking the Database Is Populated**

    A method for checking if the database has been populated already must be implemented as well. The easiest way to
    implement this is to check that there's a non-zero count of whatever the most important model in the database is.

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME
        from .models import Base

        class Manager(AbstractManager):
            module_name = MODULE_NAME

            @property
            def _base(self):
                return Base

            def populate(self):
                ...

            def is_populated(self)
                return 0 < self.session.query(MyImportantModel).count()


    **Preparing Flask Admin (Optional)**

    This class also contains a function to build a :mod:`flask` application for easy viewing of the contents of the
    database. Besides installing the optional requirements with ``python3 -m pip install flask flask-admin``, all
    that's necessary to make this available is to override the class variable ``flask_admin_models``.

    .. code-block:: python

        from bio2bel import AbstractManager
        from .constants import MODULE_NAME
        from .models import Base, Evidence, Interaction, Mirna, Species, Target

        class Manager(AbstractManager):
            module_name = MODULE_NAME
            flask_admin_models = [Evidence, Interaction, Mirna, Species, Target]

            @property
            def _base(self):
                return Base

            def populate(self):
                ...

    **Exporting to BEL (Optional)**

    If a function named ``to_bel`` is implemented that returns a :class:`pybel.BELGraph`, then the manager and CLI
    will have access to several other functions that would rely on this.
    """

    def __init__(self, connection=None, check_first=True):
        """
        :param Optional[str] connection: SQLAlchemy connection string
        :param bool check_first: Defaults to True, don't issue CREATEs for tables already present
         in the target database. Defers to :meth:`bio2bel.abstractmanager.AbstractManager.create_all`
        """
        super().__init__(connection=connection)
        self.create_all(check_first=check_first)

    @property
    @abstractmethod
    def _base(self):
        """Returns the declarative base.

        It is usually sufficient to return an instance that is module-level.

        :rtype: sqlalchemy.ext.declarative.api.DeclarativeMeta

        How to build an instance of :class:`sqlalchemy.ext.declarative.api.DeclarativeMeta` by using
        :func:`sqlalchemy.ext.declarative.declarative_base`:

        >>> from sqlalchemy.ext.declarative import declarative_base
        >>> Base = declarative_base()

        Then just override this abstract property like:

        >>> @property
        >>> def _base(self):
        >>>     return Base

        Note that this property could effectively also be a static method.
        """

    @abstractmethod
    def is_populated(self):
        """Checks if the database is already populated.

        :rtype: bool
        """

    @abstractmethod
    def populate(self, *args, **kwargs):
        """Populate the database."""

    @property
    def _metadata(self):
        """Returns the metadata object associated with this manager's declarative base."""
        return self._base.metadata

    def create_all(self, check_first=True):
        """Create the empty database (tables).

        :param bool check_first: Defaults to True, don't issue CREATEs for tables already present
         in the target database. Defers to :meth:`sqlalchemy.sql.schema.MetaData.create_all`
        """
        self._metadata.create_all(self.engine, checkfirst=check_first)

    def drop_all(self, check_first=True):
        """Drops all tables from the database.

        :param bool check_first: Defaults to True, only issue DROPs for tables confirmed to be
          present in the target database. Defers to :meth:`sqlalchemy.sql.schema.MetaData.drop_all`
        """
        self._metadata.drop_all(self.engine, checkfirst=check_first)
        self._store_drop()

    @classmethod
    def ensure(cls, connection=None):
        """Build a manager from a string, pass through a pre-built manager, or build the default manager.

        This function is a polymorphic constructor inspired by the
        `Factory Method <https://en.wikipedia.org/wiki/Factory_method_pattern>`_

        :param connection: can be either a already build manager or a connection string to build a manager with.
        :type connection: Optional[str or AbstractManager]
        :rtype: AbstractManager
        """
        if connection is None or isinstance(connection, str):
            return cls(connection=connection)

        if isinstance(connection, cls):
            return connection

        raise TypeError('passed invalid type: {}'.format(connection.__class__.__name__))



# -*- coding: utf-8 -*-

"""Aggregate CLI for all Bio2BEL projects."""

import importlib
import logging
import sys
import time

import click
from pkg_resources import VersionConflict, iter_entry_points

from .constants import DEFAULT_CACHE_CONNECTION
from .models import Action

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

cli_modules = {}
main_commands = {}
deploy_commands = {}
populate_commands = {}
drop_commands = {}
web_commands = {}

for entry_point in iter_entry_points(group='bio2bel', name=None):
    entry = entry_point.name

    try:
        bio2bel_module = entry_point.load()
    except VersionConflict:
        log.warning('Version conflict in %s', entry)
        continue
    except Exception:
        log.warning('Issue with importing module %s', entry)
        continue

    try:
        cli_modules[entry] = bio2bel_module.cli
    except AttributeError:
        try:
            cli_modules[entry] = importlib.import_module('bio2bel_{}.cli'.format(entry))
        except ImportError:
            log.warning('no submodule bio2bel_%s.cli', entry)
            continue

    try:
        main_commands[entry] = cli_modules[entry].main
    except NameError:
        log.warning('no command group bio2bel_%s.cli:main', entry)
        continue

    try:
        deploy_commands[entry] = cli_modules[entry].deploy
    except AttributeError:
        log.debug('no command bio2bel_%s.cli:deploy', entry)

    try:
        populate_commands[entry] = cli_modules[entry].populate
    except AttributeError:
        log.debug('no command bio2bel_%s.cli:populate', entry)

    try:
        drop_commands[entry] = cli_modules[entry].drop
    except AttributeError:
        log.debug('no command bio2bel_%s.cli:drop', entry)

    try:
        web_commands[entry] = cli_modules[entry].web
    except AttributeError:
        log.debug('no command bio2bel_%s.cli:web', entry)

main = click.Group(commands=main_commands)
main.help = "Bio2BEL Command Line Utilities on {}".format(sys.executable)


@main.command(help='Populate: {}'.format(', '.join(sorted(populate_commands))))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
@click.pass_context
def populate(ctx, skip):
    """Run all populate commands."""
    skip = set(ctx.params.pop('skip')) if 'skip' in ctx.params else set()

    for idx, (name, command) in enumerate(sorted(populate_commands.items()), start=1):
        if name in skip:
            click.echo('skipping {}'.format(name))
            continue

        click.echo('{} [{}/{}] populating {}'.format(time.strftime('%H:%M'), idx, len(populate_commands), name))

        try:
            command.invoke(ctx)
        except Exception:
            log.exception('error during population of %s', name)
            continue

        click.echo('{} finished populating {}'.format(time.strftime('%H:%M'), name))


@main.command(help='Drop: {}'.format(', '.join(sorted(drop_commands))))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
@click.pass_context
def drop(ctx, skip):
    """Run all drop commands."""
    skip = set(ctx.params.pop('skip')) if 'skip' in ctx.params else set()

    for name, command in sorted(drop_commands.items()):
        if name in skip:
            click.echo('skipping {}'.format(name))
            continue

        click.echo('dropping {}'.format(name))
        command.invoke(ctx)


@main.command(help='Deploy: {}'.format(', '.join(sorted(deploy_commands))))
@click.option('-s', '--skip', multiple=True, help='Modules to skip. Can specify multiple.')
@click.pass_context
def deploy(ctx, skip):
    """Run all deploy commands."""
    skip = set(ctx.params.pop('skip')) if 'skip' in ctx.params else set()

    for name, command in sorted(deploy_commands.items()):
        if name in skip:
            click.echo('skipping {}'.format(name))
            continue

        click.echo('deploying {}'.format(name))
        command.invoke(ctx)


@main.command()
@click.option('-c', '--connection', help='Defaults to {}'.format(DEFAULT_CACHE_CONNECTION))
def web(connection):
    """Run a combine web interface."""
    from bio2bel.web.application import create_application
    app = create_application(connection=connection)
    app.run(host='0.0.0.0', port=5000)


@main.command()
def web_registered():
    """Print the registered web services."""
    from bio2bel.web.application import web_modules, add_admins
    click.echo('Web Modules:')
    for manager_name in sorted(web_modules):
        click.echo(manager_name)

    click.echo('Web Admin Interfaces:')
    for manager_name, add_admin in sorted(add_admins.items()):
        click.echo('{} - {}'.format(manager_name, add_admin))


@main.command()
def actions():
    """List actions."""
    for action in Action.ls():
        click.echo('{} {} {}'.format(action.created, action.action, action.resource))


if __name__ == '__main__':
    main()
